#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Bit field inside register
"""

from . import utils
from . import config


class BitField():
    """Bit field.

    :param name: Bit field name
    :type name: str
    :param description: Bit field description
    :type description: str
    :param reset: Bit field reset vaue
    :type reset: int
    :param width: Bit field width vaue
    :type width: int
    :param lsb: Bit field LSB vaue
    :type lsb: int
    :param access: Bit field access mode
    :type access: str
    :param hardware: Bit field hardware options
    :type hardware: str
    """

    def __init__(self, name='val', description='Value of the register', reset=0, width=1,
                 lsb=0, access='rw', hardware='n', **args):
        self._enums = []

        self.name = name
        self.description = description
        self.reset = utils.str2int(reset)
        self.width = utils.str2int(width)
        self.lsb = utils.str2int(lsb)
        self.access = access
        self.hardware = hardware
        self.etc = args

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'BitField(%s)' % repr(self.name)

    def __str__(self):
        return self.as_str()

    def __len__(self):
        """Calculate number of enums inside bitfield"""
        return len(self._enums)

    def __iter__(self):
        """Create iterator over enums"""
        return iter(self._enums)

    def __getitem__(self, key):
        """Get enum by name or index."""
        try:
            if utils.is_str(key):
                key = utils.force_name_case(key)
                return next(enum for enum in self if enum.name == key)
            else:
                return self._enums[key]
        except (StopIteration, TypeError, KeyError, IndexError):
            raise KeyError("There is no enum with the name/index '%s' in the '%s' bitfield!" % (key, self.name))

    def __setitem__(self, key, value):
        """Set enum by key"""
        raise KeyError("Not able to set '%s' enum directly in the '%s' bit field!"
                       " Try to use add_enums() method." % (key, self.name))

    def as_str(self, indent=''):
        """Create an indented string with the bit field information."""
        inner_indent = indent + '  '
        bf_str = indent + '%s: %s\n' % (self.name, self.description)
        bf_str += inner_indent + 'reset = %s\n' % utils.int2str(self.reset)
        bf_str += inner_indent + 'width = %s\n' % self.width
        bf_str += inner_indent + 'lsb = %s\n' % self.lsb
        bf_str += inner_indent + 'access = %s\n' % self.access
        bf_str += inner_indent + 'hardware = %s\n' % self.hardware
        bf_str += inner_indent + 'enums:\n'
        enums = [enum.as_str(inner_indent + inner_indent) for enum in self.enums]
        bf_str += '\n'.join(enums) if enums else inner_indent + 'empty'
        return bf_str

    def as_dict(self):
        """Create a dictionary with the key attributes of the bit field."""
        d = {
            'name': self.name,
            'description': self.description,
            'reset': self.reset,
            'width': self.width,
            'lsb': self.lsb,
            'access': self.access,
            'hardware': self.hardware,
            'enums': [enum.as_dict() for enum in self.enums]
        }
        d.update(self.etc)
        return d

    @property
    def name(self):
        """Name of the bit field."""
        return utils.force_name_case(self._name)

    @name.setter
    def name(self, value):
        if not utils.is_str(value):
            raise ValueError("'name' attribute has to be 'str', but '%s' provided for the bitfield!" % type(value))
        self._name = value

    @property
    def description(self):
        """Description of the bitfield."""
        return self._description

    @description.setter
    def description(self, value):
        if not utils.is_str(value):
            raise ValueError("'description' attribute has to be 'str', but '%s' provided for '%s' bitfield!" %
                             (type(value), self.name))
        self._description = value

    @property
    def reset(self):
        """Inital value of the bitfield."""
        return self._reset

    @reset.setter
    def reset(self, value):
        self._reset = utils.str2int(value)

    @property
    def lsb(self):
        """LSB value of the bitfield."""
        return self._lsb

    @lsb.setter
    def lsb(self, value):
        self._lsb = utils.str2int(value)

    @property
    def width(self):
        """Width value of the bitfield."""
        return self._width

    @width.setter
    def width(self, value):
        self._width = utils.str2int(value)

    @property
    def access(self):
        """Access mode for the bitfield."""
        return self._access

    @access.setter
    def access(self, value):
        if not utils.is_str(value):
            raise ValueError("'access' attribute has to be 'str', but '%s' provided for '%s' bitfield!" %
                             (type(value), self.name))
        self._access = value

    @property
    def hardware(self):
        """Hardware mode for the bitfield."""
        return self._hardware

    @hardware.setter
    def hardware(self, value):
        value = value.lower()
        for ch in value:
            if ch not in "iocselaqfn":
                raise ValueError("Unknown attribute '%s' for 'hardware' property of the '%s' bitfield!" %
                                 (ch, self.name))
        self._hardware = value

    @property
    def msb(self):
        """Position of most significant bit (MSB) of the field."""
        return self.lsb + self.width - 1

    @property
    def byte_strobes(self):
        """Dictionary with LSB and MSB values for the every byte in the write data bus."""
        strb = {}
        first = self.lsb // 8
        last = self.msb // 8
        for i in range(first, last + 1):
            # per every byte strobe
            wdata_lsb = self.lsb if i == first else i * 8
            wdata_msb = (i + 1) * 8 - 1 if ((i + 1) * 8 - 1 - self.msb) < 0 else self.msb
            bf_lsb = wdata_lsb - self.lsb
            bf_msb = wdata_msb - self.lsb
            strb[i] = {'bf_lsb': bf_lsb, 'bf_msb': bf_msb,
                       'wdata_lsb': wdata_lsb, 'wdata_msb': wdata_msb}
        return strb

    @property
    def enum_names(self):
        """List with all enum names."""
        return [enum.name for enum in self]

    @property
    def enums(self):
        """List with enum objects."""
        return self._enums

    def add_enums(self, new_enums):
        """Add enum or list of enums.

        Enums are automatically sorted and stored in the ascending order of value attributes.
        """
        # hack to handle single elements
        new_enums = utils.listify(new_enums)

        # add enums to list one by one
        for enum in new_enums:
            # check existance
            assert enum.name not in self.enum_names, \
                "Enum with name '%s' is already present in '%s' bitfield!" % (enum.name, self.name)
            # check enum value is unique
            assert enum.value not in [enum.value for enum in self], \
                "Enum with value '%d' is already present in '%s' bitfield!" % (enum.value, self.name)
            # check enum conflicts with bitfield width
            assert enum.value.bit_length() <= self.width, \
                "Enum '%s' value %d exceeds bitfield width %d!" % (enum.name, enum.value, self.width)
            # if we are here - all is ok and enum can be added
            try:
                # find position to insert enum and not to break ascending order of enum values
                enum_idx = next(i for i, old_enum in enumerate(self._enums) if old_enum.value > enum.value)
                self._enums.insert(enum_idx, enum)
            except StopIteration:
                # when enum list is empty or all enum values are less than the current one
                self._enums.append(enum)
        return self

    @property
    def bits(self):
        """Create list with all positions of the bits represented by the bit field."""
        return list(range(self.lsb, self.msb + 1))

    @property
    def mask(self):
        """Bit mask for the field."""
        return ((2**(self.width) - 1) << self.lsb)

    def is_vector(self):
        """Check if the width of the bit field > 1."""
        return True if self.width > 1 else False

    def validate(self):
        """Validate parameters of the bit field."""
        # name
        assert self.name, "Empty bitfield name is not allowed!"
        assert utils.is_first_letter(self.name), \
            "Name value '%s' for is wrong! Must start from a letter." % (self.name)

        # reset
        assert utils.is_non_neg_int(self.reset), \
            "Reset value '%s' for '%s' is wrong! Only non-negative integers are allowed." % (self.reset, self.name)

        # width
        assert self.width, "Empty bitfield width is not allowed!"
        assert utils.is_pos_int(self.width), \
            "Width value '%s' for '%s' is wrong! Only positive integers are allowed." % (self.width, self.name)

        # lsb
        assert utils.is_non_neg_int(self.lsb), \
            "LSB value '%s' for '%s' is wrong! Only non-negative integers are allowed." % (self.lsb, self.name)

        # access
        assert self.access in ['rw', 'rw1c', 'rw1s', 'rw1t', 'ro', 'roc', 'roll', 'rolh', 'wo', 'wosc'], \
            "Unknown access mode '%s' for '%s' field!" % (self.access, self.name)

        # hardware
        if 'q' in self.hardware or 'n' in self.hardware or 'f' in self.hardware:
            assert len(self.hardware) == 1, \
                "Wrong hardware mode '%s' for field '%s'!" % (self.hardware, self.name)
        else:
            wrong_hw = list(set(self.hardware) - (set(self.hardware) & set('ioecsla')))
            assert wrong_hw == [], \
                "Wrong hardware mode(s) '%s' in '%s' for the field '%s'!" % (wrong_hw, self.hardware, self.name)
        if 'q' in self.hardware:
            q_access_allowed = ['rw', 'ro', 'wo']
            assert self.access in q_access_allowed, \
                "Hardware mode 'q' is allowed to use only with '%s'!" % (q_access_allowed)

        # enums
        for enum in self.enums:
            assert enum.value.bit_length() <= self.width, \
                "Enum '%s' value %d exceeds bitfield width %d!" % (enum.name, enum.value, self.width)
            assert self.enum_names.count(enum.name) == 1, \
                "Enum '%s' name is not unique!" % (enum.name)
            assert [e.value for e in self].count(enum.value) == 1, \
                "Enum '%s' value is not unique!" % (enum.value)
            enum.validate()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Utility functions and classes
"""

from . import generators
import os
from . import config
from .reg import Register
from .enum import EnumValue
from .bitfield import BitField
from .regmap import RegisterMap
from pathlib import Path


def str2int(val, base=None):
    """String to integer conversion"""
    try:
        if isinstance(val, int) or val is None:
            return val
        elif base:
            return int(val, base)
        elif '0x' in val:
            return int(val, 16)
        elif '0b' in val:
            return int(val, 2)
        else:
            return int(val)
    except (ValueError, TypeError) as e:
        raise ValueError("Can't convert '%s' to int!" % val)


def str2bool(val):
    """String to integer conversion"""
    try:
        if isinstance(val, bool):
            return val
        elif not isinstance(val, str):
            return bool(val)
        elif val.lower() in ["true", "t", "yes", "y"]:
            return True
        elif val.lower() in ["false", "f", "no", "n"]:
            return False
        else:
            raise ValueError
    except (ValueError, AttributeError, TypeError) as e:
        raise ValueError("Can't convert '%s' to bool!" % val)


def int2str(val, max_dec=1024):
    """Make string from int. Hexademical representaion will be used if input value greater that 'max_dec'."""
    if val > max_dec:
        return "0x%x" % val
    else:
        return "%d" % val


def is_non_neg_int(val):
    """Check if value is non negative integer"""
    return isinstance(val, int) and val >= 0


def is_pos_int(val):
    """Check if value is positive integer"""
    return isinstance(val, int) and val > 0


def is_str(val):
    """Check if value is string"""
    return isinstance(val, str)


def is_list(val):
    """Check if value is list"""
    return isinstance(val, list)


def is_first_letter(val):
    """Check if string starts from a letter"""
    return ord(val[0].lower()) in range(ord('a'), ord('z') + 1)


def listify(obj):
    """Make lists from single objects. No changes are made for the argument of the 'list' type."""
    if is_list(obj):
        return obj
    else:
        return [obj]


def get_file_ext(path):
    _, ext = os.path.splitext(path)
    return ext.lower()


def get_file_name(path):
    return Path(path).stem


def create_dirs(path):
    dirs = Path(path).parent
    Path(dirs).mkdir(parents=True, exist_ok=True)


def force_name_case(name):
    if config.globcfg["force_name_case"] == "upper":
        return name.upper()
    elif config.globcfg["force_name_case"] == "lower":
        return name.lower()
    else:
        return name


def create_template_simple():
    """Generate simple register map template"""
    rmap = RegisterMap()

    rmap.add_registers(Register('DATA', 'Data register', 0x0).add_bitfields(
        BitField(width=32, access='rw', hardware='ioe')))

    rmap.add_registers(Register('CTRL', 'Control register', 0x4).add_bitfields(
        BitField(width=16, access='rw', reset=0x0100, hardware='o')))

    rmap.add_registers(Register('STATUS', 'Status register', 0x8).add_bitfields(
        BitField(width=8, access='ro', hardware='i')))

    rmap.add_registers(Register('START', 'Start register', 0x100).add_bitfields(
        BitField(width=1, access='wosc', hardware='o')))

    return rmap


def create_template():
    """Generate register map template"""
    # register map
    rmap = RegisterMap()

    rmap.add_registers(Register('DATA', 'Data register', 0x4).add_bitfields([
        BitField("FIFO", "Write to push value to TX FIFO, read to get data from RX FIFO",
                 width=8, lsb=0, access='rw', hardware='q'),
        BitField("FERR", "Frame error flag. Read to clear.", width=1, lsb=16, access='rolh', hardware='i'),
        BitField("PERR", "Parity error flag. Read to clear.", width=1, lsb=17, access='rolh', hardware='i'),
    ]))

    rmap.add_registers(Register('STAT', 'Status register', 0xC).add_bitfields([
        BitField("BUSY", "Transciever is busy", width=1, lsb=2, access='ro', hardware='ie'),
        BitField("RXE", "RX FIFO is empty", width=1, lsb=4, access='ro', hardware='i'),
        BitField("TXF", "TX FIFO is full", width=1, lsb=8, access='ro', hardware='i'),
    ]))

    rmap.add_registers(Register('CTRL', 'Control register', 0x10).add_bitfields([
        BitField("BAUD", "Baudrate value", width=2, lsb=0, access='rw', hardware='o').add_enums([
            EnumValue("B9600", 0, "9600 baud"),
            EnumValue("B38400", 1, "38400 baud"),
            EnumValue("B115200", 2, "115200 baud"),
        ]),
        BitField("TXEN", "Transmitter enable. Can be disabled by hardware on error.",
                 width=1, lsb=4, access='rw', hardware='oie'),
        BitField("RXEN", "Receiver enable. Can be disabled by hardware on error.",
                 width=1, lsb=5, access='rw', hardware='oie'),
        BitField("TXST", "Force transmission start", width=1, lsb=6, access='wosc', hardware='o'),
    ]))

    rmap.add_registers(Register('LPMODE', 'Low power mode control', 0x14).add_bitfields([
        BitField("DIV", "Clock divider in low power mode", width=8, lsb=0, access='rw', hardware='o'),
        BitField("EN", "Low power mode enable", width=1, lsb=31, access='rw', hardware='o'),
    ]))

    rmap.add_registers(Register('INTSTAT', 'Interrupt status register', 0x20).add_bitfields([
        BitField("TX", "Transmitter interrupt flag. Write 1 to clear.", width=1, lsb=0, access='rw1c', hardware='s'),
        BitField("RX", "Receiver interrupt. Write 1 to clear.", width=1, lsb=1, access='rw1c', hardware='s'),
    ]))

    rmap.add_registers(Register('ID', 'IP-core ID register', 0x40).add_bitfields([
        BitField("UID", "Unique ID", width=32, lsb=0, access='ro', hardware='f', reset=0xcafe0666),
    ]))

    return rmap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Corsair is a control and status register (CSR) map generator for HDL projects.
It generates HDL code, documentation and other artifacts from CSR map description file.
"""

__title__ = "corsair"
__description__ = "Control and status register (CSR) map generator for HDL projects."

try:
    from ._version import version as __version__
except (ImportError, ModuleNotFoundError) as e:
    __version__ = 'git-latest'

from . import config
from .enum import EnumValue
from .bitfield import BitField
from .reg import Register
from .regmap import RegisterMap
from . import generators



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Output file generators
"""

import os
import json
import yaml
import jinja2
from corsair import __version__
from . import utils
from . import config
from .regmap import RegisterMap
from pathlib import Path
import wavedrom


class Generator():
    """Base generator class.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    """

    def __init__(self, rmap, **args):
        self.rmap = rmap
        self.etc = args

    def _name(self):
        return self.__class__.__name__

    def generate(self):
        """Do file generation."""
        pass

    def validate(self):
        """Validate generator parameters."""
        # config
        config.validate_globcfg(config.globcfg)
        # rmap
        assert isinstance(self.rmap, RegisterMap), \
            "Register map has to be '%s', but '%s' was provided for '%s' generator!" % (
                repr(RegisterMap()), repr(self.rmap), self._name())
        self.rmap.validate()

    def make_target(self, name):
        """Dump class attributes to dictionary that can be used as target for `csrconfig` generation.

        :param name: Name of the target
        :return: Target dictionary
        """
        params = vars(self)
        params.pop('etc')
        params.pop('rmap')
        params['generator'] = self._name()
        return {name: params}


class Jinja2():
    """Basic class for rendering Jinja2 templates"""

    def render(self, template, vars, templates_path=None):
        """Render text with Jinja2.

        :param template: Jinja2 template filename
        :param vars: Dictionary with variables for Jinja2 rendering
        :param templates_path: Path to search templates. If no path provided, then internal templates will be used
        :return: String with rendered text
        """
        # prepare template
        if not templates_path:
            templates_path = str(Path(__file__).parent / 'templates')
        j2_env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=templates_path),
                                    trim_blocks=True, lstrip_blocks=True)
        j2_env.globals.update(zip=zip)
        j2_template = j2_env.get_template(template)
        # render
        return j2_template.render(vars)

    def render_to_file(self, template, vars, path, templates_path=None):
        """Render text with Jinja2 and save it to the file.

        :param template: Jinja2 template filename
        :param vars: Dictionary with variables for Jinja2 rendering
        :param path: Path to the output file
        :param templates_path: Path to search templates. If no path provided, then internal templates will be used
        """
        # render
        rendered_text = self.render(template, vars, templates_path)
        # save
        utils.create_dirs(self.path)
        with open(path, "w") as f:
            f.write(rendered_text)


class Wavedrom():
    """Basic class for rendering register images with wavedrom"""

    def draw_regs(self, imgdir, rmap):
        imgdir.mkdir(exist_ok=True)

        bits = config.globcfg['data_width']
        lanes = bits // 16 if bits > 16 else 1
        for reg in rmap:
            reg_wd = {"reg": [],
                      "options": {"bits": bits, "lanes": lanes, "fontsize": 10}}
            bit_pos = -1
            for bf in reg:
                if bit_pos == -1 and bf.lsb > 0:
                    reg_wd["reg"].append({"bits": bf.lsb})
                elif bf.lsb - bit_pos > 1:
                    reg_wd["reg"].append({"bits": bf.lsb - bit_pos - 1})
                name = bf.name
                name_max_len = 5 * bf.width
                if len(bf.name) > name_max_len:  # to prevent labels overlapping
                    name = bf.name[:name_max_len - 1] + '..'
                reg_wd["reg"].append({"name": name, "attr": bf.access, "bits": bf.width})
                bit_pos = bf.msb
            if (bits - 1) > bit_pos:
                reg_wd["reg"].append({"bits": bits - bit_pos - 1})
            wavedrom.render(json.dumps(reg_wd)).saveas(str(imgdir / ("%s.svg" % reg.name.lower())))


class Json(Generator):
    """Dump register map to a JSON file.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    """

    def __init__(self, rmap=None, path='regs.json', **args):
        super().__init__(rmap, **args)
        self.path = path

    def generate(self):
        # validate register map
        self.rmap.validate()
        # prepare data
        data = {'regmap': list(self.rmap.as_dict().values())}
        # dump
        utils.create_dirs(self.path)
        with open(self.path, 'w') as f:
            json.dump(data, f, indent=4)


class Yaml(Generator):
    """Dump register map to a YAML file.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    """

    def __init__(self, rmap=None, path='regs.yaml', **args):
        super().__init__(rmap, **args)
        self.path = path

    def generate(self):
        # validate parameters
        self.validate()
        # prepare data
        data = {'regmap': list(self.rmap.as_dict().values())}
        # dump
        utils.create_dirs(self.path)
        with open(self.path, 'w') as f:
            yaml.Dumper.ignore_aliases = lambda *args: True  # hack to disable aliases
            yaml.dump(data, f, indent=4, default_flow_style=False, sort_keys=False)


class Txt(Generator):
    """Dump register map to a text table.

    Note: only registers with single bitfield are allowed.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    """

    def __init__(self, rmap=None, path='regs.txt', **args):
        super().__init__(rmap, **args)
        self.path = path

    def generate(self):
        # validate parameters
        self.validate()
        for reg in self.rmap:
            if len(reg) > 1:
                raise ValueError("Only registers with single bitfield are allowed for %s generator." % self._name())
        # prepare template strings
        data_w = config.globcfg['data_width']
        address_w = config.globcfg['address_width']
        address_digits = address_w // 4 + (1 if address_w % 4 else 0)
        address_str = "0x%0{0}x".format(address_digits)
        reset_digits = data_w // 4 + (1 if data_w % 4 else 0)
        reset_str = "0x%0{0}x".format(reset_digits)
        row_template = "| %-{0}s | %-{1}s | %-{2}s | %-{3}s | %-{4}s | %-{5}s | %-{6}s |\n"
        # prepare table data
        row_top = ["Address", "Name", "Width", "Access", "Hardware", "Reset", "Description"]
        col_address = [address_str % reg.address for reg in self.rmap]
        col_names = self.rmap.reg_names
        col_width = ["%d" % reg.bitfields[0].width for reg in self.rmap]
        col_access = [reg.bitfields[0].access for reg in self.rmap]
        col_hardware = [reg.bitfields[0].hardware for reg in self.rmap]
        col_reset = [reset_str % reg.reset for reg in self.rmap]
        col_description = [reg.description for reg in self.rmap]
        # calculate width of the columns for pretty printing
        cols_w = [max(address_digits + 2, len(row_top[0])),
                  max(max([len(s) for s in col_names]), len(row_top[1])),
                  max(max([len(s) for s in col_width]), len(row_top[2])),
                  max(max([len(s) for s in col_access]), len(row_top[3])),
                  max(max([len(s) for s in col_hardware]), len(row_top[4])),
                  max(reset_digits + 2, len(row_top[5])),
                  max(max([len(s) for s in col_description]), len(row_top[6]))]
        row_template = row_template.format(*cols_w)
        # render
        out_lines = [row_template % tuple(row_top)]
        out_lines.append(row_template % tuple(["-" * w for w in cols_w]))
        for i in range(len(col_names)):
            out_lines.append(row_template % (col_address[i], col_names[i],
                             col_width[i], col_access[i], col_hardware[i], col_reset[i], col_description[i]))
        # save to file
        utils.create_dirs(self.path)
        with open(self.path, 'w') as f:
            f.writelines(out_lines)


class Verilog(Generator, Jinja2):
    """Create Verilog file with register map.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param read_filler: Numeric value to return if wrong address was read
    :type read_filler: int
    :param interface: Register map bus protocol. Use one of: `axil`, `apb`, `amm`, `lb`
    :type interface: str
    """

    def __init__(self, rmap=None, path='regs.v', read_filler=0, interface='axil', **args):
        super().__init__(rmap, **args)
        self.path = path
        self.read_filler = read_filler
        self.interface = interface

    def validate(self):
        super().validate()
        assert self.interface in ['axil', 'apb', 'amm', 'lb'], \
            "Unknown '%s' interface!" % (self.interface)

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'regmap_verilog.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['module_name'] = utils.get_file_name(self.path)
        j2_vars['read_filler'] = utils.str2int(self.read_filler)
        j2_vars['interface'] = self.interface
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class Vhdl(Generator, Jinja2):
    """Create VHDL file with register map.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param read_filler: Numeric value to return if wrong address was read
    :type read_filler: int
    :param interface: Register map bus protocol. Use one of: `axil`, `apb`, `amm`, `lb`
    :type interface: str
    """

    def __init__(self, rmap=None, path='regs.vhd', read_filler=0, interface='axil', **args):
        super().__init__(rmap, **args)
        self.path = path
        self.read_filler = read_filler
        self.interface = interface

    def validate(self):
        super().validate()
        assert self.interface in ['axil', 'apb', 'amm', 'lb'], \
            "Unknown '%s' interface!" % (self.interface)

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'regmap_vhdl.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['module_name'] = utils.get_file_name(self.path)
        j2_vars['read_filler'] = utils.str2int(self.read_filler)
        j2_vars['interface'] = self.interface
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class VerilogHeader(Generator, Jinja2):
    """Create Verilog header file with register map defines.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param prefix: Prefix for the all defines. If empty output file name will be used.
    :type prefix: str
    """

    def __init__(self, rmap=None, path='regs.vh', prefix="CSR", **args):
        super().__init__(rmap, **args)
        self.path = path
        self.prefix = prefix

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'verilog_header.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['prefix'] = self.prefix.upper()
        j2_vars['file_name'] = utils.get_file_name(self.path)
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class CHeader(Generator, Jinja2):
    """Create C header file with register map defines.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param prefix: Prefix for the all defines and types. If empty output file name will be used.
    :type prefix: str
    """

    def __init__(self, rmap=None, path='regs.h', prefix="CSR", **args):
        super().__init__(rmap, **args)
        self.path = path
        self.prefix = prefix

    def validate(self):
        super().validate()
        data_width_allowed = [8, 16, 32, 64]
        assert config.globcfg['data_width'] in [8, 16, 32, 64], \
            "For %s generator, global 'data_width' must be one of '%s', but current is %d" % \
            (self._name(), data_width_allowed, config.globcfg['data_width'])

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'c_header.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['prefix'] = self.prefix.upper()
        j2_vars['file_name'] = utils.get_file_name(self.path)
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class SystemVerilogPackage(Generator, Jinja2):
    """Create SystemVerilog package with register map parameters.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param prefix: Prefix for the all parameters and types. If empty output file name will be used.
    :type prefix: str
    """

    def __init__(self, rmap=None, path='regs_pkg.sv', prefix="CSR", **args):
        super().__init__(rmap, **args)
        self.path = path
        self.prefix = prefix

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'sv_package.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['prefix'] = self.prefix.upper()
        j2_vars['file_name'] = utils.get_file_name(self.path)
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class LbBridgeVerilog(Generator, Jinja2):
    """Create Verilog file with bridge to Local Bus.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param bridge_type: Bridge protocol. Use one of `axil`, `apb`, `amm`.
    :type bridge_type: str
    """

    def __init__(self, rmap=None, path='axil2lb.v', bridge_type='axil', **args):
        super().__init__(rmap, **args)
        self.path = path
        self.bridge_type = bridge_type

    def validate(self):
        assert self.bridge_type in ['axil', 'apb', 'amm'], \
            "Unknown '%s' bridge type!" % (self.bridge_type)

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        if self.bridge_type == 'axil':
            j2_template = 'axil2lb_verilog.j2'
        elif self.bridge_type == 'apb':
            j2_template = 'apb2lb_verilog.j2'
        elif self.bridge_type == 'amm':
            j2_template = 'amm2lb_verilog.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['module_name'] = utils.get_file_name(self.path)
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class LbBridgeVhdl(Generator, Jinja2):
    """Create VHDL file with bridge to Local Bus.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param bridge_type: Bridge protocol. Use one of `axil`, `apb`, `amm`.
    :type bridge_type: str
    """

    def __init__(self, rmap=None, path='axil2lb.v', bridge_type='axil', **args):
        super().__init__(rmap, **args)
        self.path = path
        self.bridge_type = bridge_type

    def validate(self):
        assert self.bridge_type in ['axil', 'apb', 'amm'], \
            "Unknown '%s' bridge type!" % (self.bridge_type)

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        if self.bridge_type == 'axil':
            j2_template = 'axil2lb_vhdl.j2'
        elif self.bridge_type == 'apb':
            j2_template = 'apb2lb_vhdl.j2'
        elif self.bridge_type == 'amm':
            j2_template = 'amm2lb_vhdl.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['module_name'] = utils.get_file_name(self.path)
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)


class Markdown(Generator, Jinja2, Wavedrom):
    """Create documentation for a register map in Markdown.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param title: Document title
    :type title: str
    :param print_images: Enable generating images for bit fields of a register
    :type print_images: bool
    :param image_dir: Path to directory where all images will be saved
    :type image_dir: str
    :param print_conventions: Enable generating table with register access modes explained
    :type print_conventions: bool
    """

    def __init__(self, rmap=None, path='regs.md', title='Register map',
                 print_images=True, image_dir="regs_img", print_conventions=True, **args):
        super().__init__(rmap, **args)
        self.path = path
        self.title = title
        self.print_images = print_images
        self.image_dir = image_dir
        self.print_conventions = print_conventions

    def generate(self):
        filename = utils.get_file_name(self.path)
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'regmap_md.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['print_images'] = utils.str2bool(self.print_images)
        j2_vars['print_conventions'] = utils.str2bool(self.print_conventions)
        j2_vars['image_dir'] = self.image_dir
        j2_vars['filename'] = filename
        j2_vars['title'] = self.title
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)
        # draw register images
        if self.print_images:
            self.draw_regs(Path(self.path).parent / self.image_dir, self.rmap)


class Asciidoc(Generator, Jinja2, Wavedrom):
    """Create documentation for a register map in AsciiDoc.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    :param title: Document title
    :type title: str
    :param print_images: Enable generating images for bit fields of a register
    :type print_images: bool
    :param image_dir: Path to directory where all images will be saved
    :type image_dir: str
    :param print_conventions: Enable generating table with register access modes explained
    :type print_conventions: bool
    """

    def __init__(self, rmap=None, path='regs.adoc', title='Register map',
                 print_images=True, image_dir="regs_img", print_conventions=True, **args):
        super().__init__(rmap, **args)
        self.path = path
        self.title = title
        self.print_images = print_images
        self.image_dir = image_dir
        self.print_conventions = print_conventions

    def generate(self):
        filename = utils.get_file_name(self.path)
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'regmap_asciidoc.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['print_images'] = utils.str2bool(self.print_images)
        j2_vars['print_conventions'] = utils.str2bool(self.print_conventions)
        j2_vars['image_dir'] = self.image_dir
        j2_vars['filename'] = filename
        j2_vars['title'] = self.title
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)
        # draw register images
        if self.print_images:
            self.draw_regs(Path(self.path).parent / self.image_dir, self.rmap)


class Python(Generator, Jinja2):
    """Create Python file to access register map via some interface.

    :param rmap: Register map object
    :type rmap: :class:`corsair.RegisterMap`
    :param path: Path to the output file
    :type path: str
    """

    def __init__(self, rmap=None, path='regs.py', **args):
        super().__init__(rmap, **args)
        self.path = path

    def generate(self):
        # validate parameters
        self.validate()
        # prepare jinja2
        j2_template = 'regmap_py.j2'
        j2_vars = {}
        j2_vars['corsair_ver'] = __version__
        j2_vars['rmap'] = self.rmap
        j2_vars['config'] = config.globcfg
        # render
        self.render_to_file(j2_template, j2_vars, self.path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Control and status register
"""

from . import utils
from . import config


class Register():
    """Control and status register.

    :param name: Bit field name
    :type name: str
    :param description: Bit field description
    :type description: str
    :param address: Register address
    :type address: int, None
    """

    def __init__(self, name='csr0', description='Control and status register 0', address=None, **args):
        self._bitfields = []

        self.name = name
        self.description = description
        self.address = address
        self.etc = args

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'Register(%s, %s, %s)' % (repr(self.name), repr(self.description), repr(self.address))

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Create an indented string with the information about the register."""
        inner_indent = indent + '  '
        bitfields = [bf.as_str(inner_indent) for bf in self.bitfields]
        bitfields_str = '\n'.join(bitfields) if bitfields else inner_indent + 'empty'
        reg_str = indent + '(0x%x) %s: %s\n' % (self.address, self.name, self.description)
        reg_str += bitfields_str
        return reg_str

    def as_dict(self):
        """Create a dictionary with the key attributes of the register."""
        d = {
            'name': self.name,
            'description': self.description,
            'address': self.address,
            'bitfields': [bf.as_dict() for bf in self.bitfields]
        }
        d.update(self.etc)
        return d

    def __len__(self):
        """Calculate number of bit fields inside register"""
        return len(self._bitfields)

    def __iter__(self):
        """Create iterator over bit fields """
        return iter(self._bitfields)

    def __getitem__(self, key):
        """Get bit field by name or index"""
        try:
            if utils.is_str(key):
                key = utils.force_name_case(key)
                return next(bf for bf in self if bf.name == key)
            else:
                return self._bitfields[key]
        except (StopIteration, TypeError, KeyError, IndexError):
            raise KeyError("There is no bit field with the name/index '%s' in the '%s' register!" % (key, self.name))

    def __setitem__(self, key, value):
        """Set bit field by key"""
        raise KeyError("Not able to set '%s' bit field directly in the '%s' register!"
                       " Try to use add_bitfields() method." % (key, self.name))

    @property
    def name(self):
        """Name of the register."""
        return utils.force_name_case(self._name)

    @name.setter
    def name(self, value):
        if not utils.is_str(value):
            raise ValueError("'name' attribute has to be 'str', but '%s' provided for the register!" % type(value))
        self._name = value

    @property
    def address(self):
        """Address of the register."""
        return self._address

    @address.setter
    def address(self, value):
        self._address = utils.str2int(value)

    @property
    def description(self):
        """Description of the register."""
        return self._description

    @description.setter
    def description(self, value):
        if not utils.is_str(value):
            raise ValueError("'description' attribute has to be 'str', but '%s' provided for '%s' register!" %
                             (type(value), self.name))
        self._description = value

    @property
    def bitfield_names(self):
        """List with all bit field names."""
        return [bf.name for bf in self]

    @property
    def bitfields(self):
        """List with bit field objects."""
        return self._bitfields

    def add_bitfields(self, new_bitfields):
        """Add bit field or list of bit feilds.

        Bit fields are automatically sorted and stored in the ascending order of msb attributes.
        """
        # hack to handle single elements
        new_bitfields = utils.listify(new_bitfields)

        # add bit fields to list one by one
        for bf in new_bitfields:
            # check existance
            assert bf.name not in self.bitfield_names, \
                "Bit field with name '%s' is already present in '%s' register!" % (bf.name, self.name)
            # check fields overlapping
            overlaps = [set(bf.bits).intersection(set(old_bf.bits)) for old_bf in self._bitfields]
            overlaps_names = [self.bitfield_names[i] for i, ovl in enumerate(overlaps) if ovl]
            assert not (self.bitfields and overlaps_names), \
                "Position of a bit field '%s' conflicts with other bit field(s): %s!" % (bf.name, repr(overlaps_names))
            # check bit field conflicts with data width
            data_width = config.globcfg['data_width']
            assert bf.msb < data_width, \
                "Field '%s' (msb=%d) exceeds interface data width %d!" % \
                (bf.name, bf.msb, data_width)
            # if we are here - all is ok and bit field can be added
            try:
                # find position to insert bit field and not to break ascending order of bit field msb positions
                bf_idx = next(i for i, old_bf in enumerate(self._bitfields) if old_bf.msb > bf.msb)
                self._bitfields.insert(bf_idx, bf)
            except StopIteration:
                # when bit field list is empty or all bit field msb positions are less than the current one
                self._bitfields.append(bf)
        return self

    @property
    def reset(self):
        """Reset value of the refister after reset."""
        init = 0
        for bf in self:
            init |= bf.reset << bf.lsb
        return init

    @property
    def access(self):
        """Register access mode, based on bitfields."""
        accesses = list(set([bf.access for bf in self.bitfields]))
        if len(accesses) == 1:
            return accesses[0][:2]
        else:
            return 'rw'

    def validate(self):
        """Validate parameters of the register."""
        # name
        assert self.name, "Empty register name is not allowed!"
        assert utils.is_first_letter(self.name), \
            "'name' value '%s' for the register is wrong! Must start from a letter." % (self.name)

        # address
        assert utils.is_non_neg_int(self.address), \
            "Address value '%s' for '%s' is wrong! Only non-negative integers are allowed." % (self.address, self.name)

        # bit fields overlapping
        for bf in self._bitfields:
            overlaps = [set(bf.bits).intersection(set(bf_.bits)) for bf_ in self._bitfields]
            overlaps_names = [self.bitfield_names[i] for i, ovl in enumerate(overlaps) if ovl]
            overlaps_names.pop(overlaps_names.index(bf.name))
            assert not overlaps_names, \
                "Position and size of a bit field '%s' conflicts with other bit field(s): %s!" % \
                (bf.name, repr(overlaps_names))

        # bit fields vs data_width
        data_width = config.globcfg['data_width']
        for bf in self._bitfields:
            assert bf.msb < data_width, \
                "Field '%s' (msb=%d) exceeds interface data width %d!" % \
                (bf.name, bf.msb, data_width)

        # bit fields
        for bf in self.bitfields:
            assert self.bitfield_names.count(bf.name) == 1, \
                "Bitfield '%s' name is not unique!" % (bf.name)
            bf.validate()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Module to operate with Corsair configuration files
"""

from . import utils
import configparser


def default_globcfg():
    """Create a global configuration with parameters by default."""
    return {
        "base_address": 0,
        "data_width": 32,
        "address_width": 16,
        "register_reset": "sync_pos",
        "address_increment": "none",
        "address_alignment": "data_width",
        "force_name_case": "none",
    }


def read_csrconfig(cfgpath):
    """Parse Corsair configuration file. Return two dictionaries: global configuration and targets."""
    # open config file
    cfg = configparser.ConfigParser()
    cfg.read_file(open(cfgpath, 'r'))

    # parse global config
    globcfg = default_globcfg()
    if 'globcfg' in cfg.sections():
        globcfg.update(dict(cfg['globcfg']))
        globcfg["base_address"] = utils.str2int(globcfg["base_address"])
        globcfg["data_width"] = utils.str2int(globcfg["data_width"])
        globcfg["address_width"] = utils.str2int(globcfg["address_width"])
        try:
            globcfg["address_increment"] = utils.str2int(globcfg["address_increment"])
        except ValueError:
            pass
        try:
            globcfg["address_alignment"] = utils.str2int(globcfg["address_alignment"])
        except ValueError:
            pass
        validate_globcfg(globcfg)

    # parse targets
    targets = {}
    for target_name in [name for name in cfg.sections() if name != 'globcfg']:
        if 'generator' in dict(cfg[target_name]).keys():
            targets[target_name] = dict(cfg[target_name])

    return globcfg, targets


def write_csrconfig(cfgpath, globcfg, targets):
    """Save Corsair configuration file."""
    cfg = configparser.ConfigParser()
    data = {'globcfg': globcfg}
    data.update(targets)
    cfg.read_dict(data)
    cfg.write(open(cfgpath, 'w'))


def validate_globcfg(globcfg):
    """Validate a dictionary with global configuration."""
    # base_address
    assert utils.is_non_neg_int(globcfg["base_address"]), \
        "Wrong value for 'base_address'='%s'. Must be a non negative integer." % globcfg["base_address"]

    # data_width
    assert utils.is_non_neg_int(globcfg["data_width"]), \
        "Wrong value for 'data_width'='%s'. Must be a non negative integer." % globcfg["data_width"]

    # address_width
    assert utils.is_non_neg_int(globcfg["address_width"]), \
        "Wrong value for 'address_width'='%s'. Must be a non negative integer." % globcfg["address_width"]

    # register_reset
    register_reset_allowed = ['sync_pos', 'sync_neg', 'async_pos', 'async_neg']
    assert globcfg["register_reset"] in register_reset_allowed, \
        "Wrong value for 'register_reset'='%s'. Must be one of this: %s." % (globcfg["address_width"],
                                                                             register_reset_allowed)

    # address_increment
    address_increment_alowed = ['none', 'data_width']
    try:
        is_valid = (globcfg["address_increment"] in address_increment_alowed or
                    utils.is_non_neg_int(globcfg["address_increment"]))
    except ValueError:
        is_valid = False
    assert is_valid, \
        "Wrong value for 'address_increment'='%s'. Must be one of this: %s or a non negative integer." % (
            globcfg["address_increment"], address_increment_alowed)

    # address_alignment
    address_alignment_alowed = ['none', 'data_width']
    try:
        is_valid = (globcfg["address_alignment"] in address_alignment_alowed or
                    utils.is_non_neg_int(globcfg["address_alignment"]))
    except ValueError:
        is_valid = False
    assert is_valid, \
        "Wrong value for 'address_alignment'='%s'. Must be one of this: %s or a non negative integer." % (
            globcfg["address_alignment"], address_alignment_alowed)

    # force_name_case
    force_name_case_allowed = ['lower', 'upper', 'none']
    assert globcfg["force_name_case"] in force_name_case_allowed, \
        "Wrong value for 'force_name_case'='%s'. Must be one of this: %s." % (globcfg["force_name_case"],
                                                                              force_name_case_allowed)


globcfg = default_globcfg()


def set_globcfg(globcfg_):
    """Use specified global configuration for all operations"""
    global globcfg
    validate_globcfg(globcfg_)
    globcfg = globcfg_



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Enumerated value for bit field
"""

from . import utils
from . import config


class EnumValue():
    """Enumerated value.

    :param name: Enum name
    :type name: str
    :param value: Enum value
    :type value: int
    :param description: Enum description
    :type description: str
    """

    def __init__(self, name='enum', value=0, description='Enumerated value', **args):
        self.name = name
        self.description = description
        self.value = value
        self.etc = args

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" %
                            (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" %
                            (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'EnumValue(%s)' % repr(self.name)

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Create an indented string with the enum information."""
        enum_str = indent + \
            '%s: %d - %s' % (self.name, self.value, self.description)
        return enum_str

    def as_dict(self):
        """Create a dictionary with the key attributes of the bit field."""
        d = {'name': self.name, 'description': self.description, 'value': self.value}
        d.update(self.etc)
        return d

    @property
    def name(self):
        """Name of the enum."""
        return utils.force_name_case(self._name)

    @name.setter
    def name(self, value):
        if not utils.is_str(value):
            raise ValueError(
                "'name' attribute has to be 'str', but '%s' provided for the enum!" % type(value))
        self._name = value

    @property
    def value(self):
        """Value of the enum."""
        return self._value

    @value.setter
    def value(self, value_):
        self._value = utils.str2int(value_)

    @property
    def description(self):
        """Description of the enum."""
        return self._description

    @description.setter
    def description(self, value):
        if not utils.is_str(value):
            raise ValueError("'description' attribute has to be 'str', but '%s' provided for the '%s' enum!" %
                             (type(value), self.name))
        self._description = value

    def validate(self):
        """Validate parameters of the enum."""
        # value type
        assert utils.is_non_neg_int(self.value), \
            "Enum value '%s' for '%s' enum is wrong! Only non-negative integers are allowed." % \
            (self.value, self.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Register map
"""

from . import utils
from . import config
from .reg import Register
from .bitfield import BitField
from .enum import EnumValue
import json
import yaml


class RegisterMap():
    """CSR map"""

    def __init__(self):
        self._regs = []

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return self.as_dict() == other.as_dict()

    def __ne__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError("Failed to compare '%s' with '%s'!" % (repr(self), repr(other)))
        else:
            return not self.__eq__(other)

    def __repr__(self):
        return 'RegisterMap()'

    def __str__(self):
        return self.as_str()

    def as_str(self, indent=''):
        """Create indented string with the information about register map."""
        inner_indent = indent + '  '
        regs = [reg.as_str(inner_indent) for reg in self.regs]
        regs_str = '\n'.join(regs) if regs else inner_indent + 'empty'
        return indent + 'register map:\n' + regs_str

    def as_dict(self):
        """Return register map as a dictionary."""
        return {reg.name: reg.as_dict() for reg in self.regs}

    def __len__(self):
        """Calculate number of the registers"""
        return len(self._regs)

    def __iter__(self):
        """Create iterator over registers"""
        return iter(self._regs)

    def __getitem__(self, key):
        """Get register by name or index"""
        try:
            if isinstance(key, str):
                return next(reg for reg in self if reg.name == key)
            else:
                return self._regs[key]
        except (StopIteration, TypeError, KeyError, IndexError):
            raise KeyError("There is no register with a name/index '%s'!" % (key))

    def __setitem__(self, key, value):
        """Set register by key"""
        raise KeyError("Not able to set '%s' register directly!"
                       " Try use add_registers() method." % (key))

    @property
    def reg_names(self):
        """List with all register names."""
        return [reg.name for reg in self]

    def _addr_resolve(self, reg):
        """Resolve address for a register with no address."""
        # some error checks
        assert len(self) != 0, \
            "Register '%s' with no address is not allowed to be the first register in a map!" % (reg.name)
        assert config.globcfg['address_increment'] != 'none', \
            "Register '%s' with no address is not allowed when address auto increment is disabled!" % (reg.name)

        prev_addr = self.regs[-1].address

        if config.globcfg['address_increment'] == 'data_width':
            addr_step = config.globcfg['data_width'] // 8
        else:
            addr_step = config.globcfg['address_increment']

        reg.address = prev_addr + addr_step

    def _addr_check_alignment(self, reg):
        """Check address alignment."""
        if config.globcfg['address_alignment'] == 'none':
            align_val = 1
        elif config.globcfg['address_alignment'] == 'data_width':
            align_val = config.globcfg['data_width'] // 8
        else:
            align_val = config.globcfg['address_alignment']

        assert (reg.address % align_val) == 0, \
            "Register '%s' with address '%d' is not %d bytes alligned!" % (reg.name, reg.address, align_val)

    def _addr_check_conflicts(self, reg):
        addresses = [reg.address for reg in self]
        if reg.address in addresses:
            conflict_reg = self[addresses.index(reg.address)]
            assert False, "Register '%s' with address '%d' conflicts with register '%s' with the same address!" % \
                (reg.name, reg.address, conflict_reg.name)

    @property
    def regs(self):
        """List with register objects."""
        return self._regs

    def add_registers(self, new_regs):
        """Add list of registers.

        Register are automatically sorted and stored in the ascending order of addresses.
        """
        # hack to handle single elements
        new_regs = utils.listify(new_regs)

        # add registers to the list one by one
        for reg in new_regs:
            # check existance
            assert reg.name not in self.reg_names, \
                "Register with name '%s' is already present!" % (reg.name)
            # aplly calculated address if register address is empty
            if reg.address is None:
                self._addr_resolve(reg)
            # check address alignment
            self._addr_check_alignment(reg)
            # check address conflicts
            self._addr_check_conflicts(reg)
            # if we here - all is ok and register can be added
            try:
                # find position to insert register and not to break ascending order of addresses
                reg_idx = next(i for i, r in enumerate(self._regs) if r.address > reg.address)
                self._regs.insert(reg_idx, reg)
            except StopIteration:
                # when registers list is empty or all addresses are less than the current one
                self._regs.append(reg)
        return self

    def validate(self):
        """Validate the register map."""
        for reg in self.regs:
            assert self.reg_names.count(reg.name) == 1, \
                "Register '%s' name is not unique!" % (reg.name)
            reg.validate()

    def read_file(self, path):
        """Read register map from file (based on extension)."""
        ext = utils.get_file_ext(path)
        if ext in ['.yaml', '.yml']:
            self.read_yaml(path)
        elif ext == '.json':
            self.read_json(path)
        elif ext == '.txt':
            self.read_txt(path)
        else:
            raise ValueError("Unknown extension '%s' of the file '%s'" % (ext, path))

    def read_json(self, path):
        """Read register map from JSON file."""
        with open(path, 'r') as f:
            data = json.load(f)
            self._fill_from_file_data(data['regmap'])

    def read_yaml(self, path):
        """Read register map from YAML file."""
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
            self._fill_from_file_data(data['regmap'])

    def read_txt(self, path):
        """Read register map from text file."""
        with open(path, 'r') as f:
            raw_lines = f.readlines()
            data = []
            reg_start_idx = None
            for i, line in enumerate(raw_lines):
                if '-----' in line:
                    reg_start_idx = i + 1
                if reg_start_idx and i >= reg_start_idx:
                    # register with one bitfield template
                    reg = {"name": None, "description": None,
                           "bitfields": [{"width": None, "access": None, "hardware": None}]}
                    # prepare the line
                    line_data = [s.strip() for s in line.split("|")[1:-1]]
                    if len(line_data) != 7:
                        raise ValueError("Not enough / too much columns in line %d. Plese fix!" % (i + 1))
                    # extract register properties
                    if line_data[0]:  # address value can be ommited
                        reg["address"] = line_data[0]
                    reg["address"] = line_data[0] if line_data[0] else None
                    reg["name"] = line_data[1]
                    reg["bitfields"][0]["width"] = line_data[2]
                    reg["bitfields"][0]["access"] = line_data[3]
                    reg["bitfields"][0]["hardware"] = line_data[4]
                    if line_data[4]:  # reset value can be ommited
                        reg["bitfields"][0]["reset"] = line_data[5]
                    reg["description"] = line_data[6]
                    data.append(reg)
            if not reg_start_idx:
                raise ValueError("Can't find table with registers!")
            self._fill_from_file_data(data)

    def _fill_from_file_data(self, data):
        """Fill register map with data from file."""
        self._regs = []
        for data_reg in data:
            data_reg_filtered = {k: v for k, v in data_reg.items() if k != 'bitfields'}
            reg = Register(**data_reg_filtered)
            for data_bf in data_reg['bitfields']:
                data_bf_filtered = {k: v for k, v in data_bf.items() if k != 'enums'}
                bf = BitField(**data_bf_filtered)
                if 'enums' in data_bf.keys():
                    for data_enum in data_bf['enums']:
                        bf.add_enums(EnumValue(**data_enum))
                reg.add_bitfields(bf)
            self.add_registers(reg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Run Corsair from command line with arguments.
"""

import sys
import os
import argparse
from pathlib import Path
import corsair
from . import utils
from contextlib import contextmanager
import importlib.util

__all__ = ['main']


@contextmanager
def cwd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


class ArgumentParser(argparse.ArgumentParser):
    """Inherit ArgumentParser to override the behaviour of error method."""

    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(2, '\n%s: error: %s\n' % (self.prog, message))


def parse_arguments():
    """Parse and validate arguments."""
    parser = ArgumentParser(prog=corsair.__title__,
                            description=corsair.__description__)
    parser.add_argument('-v', '--version',
                        action='version',
                        version='%(prog)s v' + corsair.__version__)
    parser.add_argument(metavar='WORKDIR',
                        nargs='?',
                        dest='workdir_path',
                        default=os.getcwd(),
                        help='working directory (default is the current directory)')
    parser.add_argument('-r',
                        metavar='REGMAP',
                        dest='regmap_path',
                        help='read register map from file')
    parser.add_argument('-c',
                        metavar='CONFIG',
                        dest='config_path',
                        help='read configuration from file')
    template_choices = ['json', 'yaml', 'txt']
    parser.add_argument('-t',
                        metavar='FORMAT',
                        choices=template_choices,
                        dest='template_format',
                        help='create templates (choose from %s)' % ', '.join(template_choices))
    return parser.parse_args()


def generate_templates(format):
    print("... templates format: '%s'" % format)
    # global configuration
    globcfg = corsair.config.default_globcfg()
    globcfg['data_width'] = 32
    globcfg['address_width'] = 16
    globcfg['register_reset'] = 'sync_pos'
    corsair.config.set_globcfg(globcfg)

    # targets
    targets = {}
    targets.update(corsair.generators.Verilog(path="hw/regs.v").make_target('v_module'))
    targets.update(corsair.generators.Vhdl(path="hw/regs.vhd").make_target('vhdl_module'))
    targets.update(corsair.generators.VerilogHeader(path="hw/regs.vh").make_target('v_header'))
    targets.update(corsair.generators.SystemVerilogPackage(path="hw/regs_pkg.sv").make_target('sv_pkg'))
    targets.update(corsair.generators.Python(path="sw/regs.py").make_target('py'))
    targets.update(corsair.generators.CHeader(path="sw/regs.h").make_target('c_header'))
    targets.update(corsair.generators.Markdown(path="doc/regs.md", image_dir="md_img").make_target('md_doc'))
    targets.update(corsair.generators.Asciidoc(path="doc/regs.adoc", image_dir="adoc_img").make_target('asciidoc_doc'))

    # create templates
    if format == 'txt':
        rmap = corsair.utils.create_template_simple()
    else:
        rmap = corsair.utils.create_template()
    # register map template
    if format == 'json':
        gen = corsair.generators.Json(rmap)
        regmap_path = 'regs.json'
    elif format == 'yaml':
        gen = corsair.generators.Yaml(rmap)
        regmap_path = 'regs.yaml'
    elif format == 'txt':
        gen = corsair.generators.Txt(rmap)
        regmap_path = 'regs.txt'
    print("... generate register map file '%s'" % regmap_path)
    gen.generate()
    # configuration file template
    config_path = 'csrconfig'
    globcfg['regmap_path'] = regmap_path
    print("... generate configuration file '%s'" % config_path)
    corsair.config.write_csrconfig(config_path, globcfg, targets)


def die(msg):
    print("Error: %s" % msg)
    exit(1)


def finish():
    print('Success!')
    exit(0)


def app(args):
    print("... set working directory '%s'" % args.workdir_path)

    # check if teplates are needed
    if args.template_format:
        generate_templates(args.template_format)
        finish()

    # check if configuration file path was provided
    if args.config_path:
        config_path = Path(args.config_path)
    else:
        config_path = Path('csrconfig')
    # check it existance
    if not config_path.is_file():
        die("Can't find configuration file '%s'!" % config_path)
    # try to read it
    print("... read configuration file '%s'" % config_path)
    globcfg, targets = corsair.config.read_csrconfig(config_path)

    # check if regiter map file path was provided
    if args.regmap_path:
        regmap_path = Path(args.regmap_path)
        globcfg['regmap_path'] = regmap_path
    elif 'regmap_path' in globcfg.keys():
        regmap_path = Path(globcfg['regmap_path'])
    else:
        regmap_path = None
        print("Warning: No register map file was specified!")
    corsair.config.set_globcfg(globcfg)

    if regmap_path:
        # check it existance
        if not regmap_path.is_file():
            die("Can't find register map file '%s'!" % regmap_path)
        # try to read it
        print("... read register map file '%s'" % regmap_path)
        rmap = corsair.RegisterMap()
        rmap.read_file(regmap_path)
        print("... validate register map")
        rmap.validate()
    else:
        rmap = None

    # make targets
    if not targets:
        die("No targets were specified! Nothing to do!")
    for t in targets:
        print("... make '%s': " % t, end='')
        if 'generator' not in targets[t].keys():
            die("No generator was specified for the target!")

        if '.py::' in targets[t]['generator']:
            custom_module_path, custom_generator_name = targets[t]['generator'].split('::')
            custom_module_name = utils.get_file_name(custom_module_path)
            spec = importlib.util.spec_from_file_location(custom_module_name, custom_module_path)
            custom_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(custom_module)
            try:
                gen_obj = getattr(custom_module, custom_generator_name)
                gen_name = custom_generator_name
            except AttributeError:
                die("Generator '%s' from module '%s' does not exist!" % (custom_generator_name, custom_module_path))
        else:
            gen_name = targets[t]['generator']
            try:
                gen_obj = getattr(corsair.generators, gen_name)
            except AttributeError:
                die("Generator '%s' does not exist!" % gen_name)

        gen_args = targets[t]
        print("%s -> '%s': " % (gen_name, gen_args['path']))
        gen_obj(rmap, **gen_args).generate()


def main():
    """Program main"""
    # parse arguments
    args = parse_arguments()

    # do all the things inside working directory
    args.workdir_path = str(Path(args.workdir_path).absolute())
    with cwd(args.workdir_path):
        app(args)
    finish()


if __name__ == '__main__':
    main()
